import os, random, time

from pathlib import Path
from typing import Any, Literal, Callable
from colorama import Style, Fore

from .location import *
from .space import *

class ScriptHandler:
    def __init__(self) -> None:
        ...
    def addClass(self, cls: type):
        self.__dict__ |= {
            f"{cls.__name__.capitalize()}Class": cls,
            cls.__name__.lower(): cls()
        }
    @staticmethod
    def createClassFromString(name: str, s: str) -> type:
        class_name: str = name.replace("-", "_")
        s = "..." if not s else s
        lns: list[str] = f"""
from brary import *
class {class_name}:
    def __getattr__(self, name: str) -> None:
        raise AttributeError(f"attempted to retrieve {name} attribute '{{name}}', but it is not defined in the {name} script.")
        """.strip().split("\n") + [f"    {ln}" for ln in s.split("\n")]
        s = "\n".join(lns)
        exec(s)
        return locals()[class_name]()
    @staticmethod
    def createClass(name: str) -> type:
        new_name: str = ""
        set_upper: bool = False
        for char in name:
            if char == " ":
                set_upper = True
            elif set_upper:
                new_name += char.upper()
                set_upper = False
            else:
                new_name += char
        name = new_name
        scripts_path: Path = Path(os.getcwd()) / "scripts"
        file_path: Path = scripts_path / name
        if not scripts_path.exists():
            raise FileNotFoundError(f"sh setup() requires a scripts directory in your current working directory ({os.getcwd()}). you can run 'mkdir {scripts_path}' to create it.")
        if not file_path.exists():
            raise FileNotFoundError(f"sh setup() requires {name} script in {scripts_path}")
        with file_path.open() as file:
            code: str = file.read()
        return ScriptHandler.createClassFromString(name, code)
class Game:
    def __init__(self, board: "Board") -> None:
        self.board: "Board" = board
        self.player: object = ScriptHandler.createClass("Player")
        self.player.inventory_enabled = True
        self.piston: Piston = board.new(Piston)
        self.tos: "list[object]" = [] # tick objects
        self.shops: "list[Shop, str, bool]" = [] # [shop, command, enabled]
    def nto(self, obj: object) -> None: # new tick object
        self.tos.append(obj)
        return obj
    def newShop(self, shop: "Shop", command: str) -> None:
        self.shops.append([shop, command, True])
    def toggleShop(self, shop: "Shop | int", to: bool | None = None) -> None: # 'to' argument toggles when set to None
        shop = self.shops[shop] if isinstance(shop, int) else shop
        for iter_shop in self.shops:
            if iter_shop[0].name == shop.name:
                iter_shop[2] = not iter_shop[2] if to is None else to
    def newInventory(self, ii_cls: type | None = None) -> None:
        from .economy import ItemInventory

        ii_cls = ii_cls or ItemInventory
        self.player.inventory = ii_cls.setup()
    def loop(self, clear: bool = True, until: Callable | bool = False, player_tag: str = "Player", *, dno_secured: bool = False) -> None:
        if not dno_secured:
            try:
                self.loop(clear, until, player_tag, dno_secured=True)
            except (KeyboardInterrupt, EOFError):
                ...
            return
        while not (until() if callable(until) else until):
            for obj in self.tos:
                obj.tick()
            if clear:
                os.system("cls" if os.name == "nt" else "clear")
            print(self.board.render())
            q: str = input("> ")
            if q in list("qweasdzxc"):
                self.piston.push(self.board.idByTag(player_tag), q, self.player.S, cannot_go_there_warning=True)
            elif q == "i":
                if not hasattr(self.player, "inventory") or not self.player.inventory_enabled:
                    print(f"{Fore.RED}inventory unavailable!{Style.RESET_ALL}")
                    time.sleep(0.4)
                    continue
                print(f"{Fore.GREEN}inventory:{Style.RESET_ALL}")
                print(f"1. {self.player.inventory.wallet}")
                print(*([f"{i}. {item.name}" for i, item in enumerate(self.player.inventory.items, start=2)] or [f"{Fore.LIGHTBLACK_EX}...nothing in inventory...{Style.RESET_ALL}"]), sep="\n")
                while True:
                    try:
                        q2: str | int = input(f"{Fore.LIGHTBLACK_EX}press <return> to close inventory or type command here: {Style.RESET_ALL}").lower()
                        if q2 == "":
                            break
                        q2 = int(q2) - 2
                        if q2 < 0:
                            continue
                    except ValueError:
                        print(f"{Fore.RED}invalid command{Style.RESET_ALL}")
                        time.sleep(0.4)
                        continue
                    if q2 < len(self.player.inventory.items):
                        print(f"{Fore.LIGHTBLACK_EX}using {self.player.inventory.items[q2].name}{Style.RESET_ALL}")
                        self.player.inventory.items[q2].use(self)
                        break
            else:
                if self.player.inventory_enabled:
                    for shop in self.shops:
                        if q == shop[1] and shop[2] is True:
                            item: "ShopItemListing" = shop[0].display()
                            if item is not None:
                                item.buy(self.player.inventory)
                            break

class Board:
    def __init__(self, size: LocationArray, bg: TileTexture) -> None:
        self.size: LocationArray = size
        self.states: list[BoardState] = []
        self.bg: TileTexture = bg
    @classmethod
    def setup(cls) -> "Board":
        sh: ScriptHandler = ScriptHandler()
        board: object = sh.createClass("Board")
        return cls(LocationArray(LocationX(board.W), LocationY(board.H)), TileTexture(board.T, board.C))
    @staticmethod
    def generateRandomSpaceId() -> int:
        return random.randint(0, 999999)
    def new(self, cls: type) -> Any:
        return cls(self)
    def stateTaken(self, location: LocationArray) -> bool:
        for state in self.states:
            if LocationArray.compare(state.location, location):
                return state
        return False
    class StatePlacementOperator:
        def __init__(
            self,
            board: "Board",
            space_id: int,
            location: LocationArray,
            tag: str,
            texture: TileTexture,
            overwrite: bool = True
        ) -> None:
            self.board: "Board" = board
            self.ensureLocationValidity(board, location)
            state: BoardState | False = board.stateTaken(location)
            if not overwrite and state is not False:
                raise ValueError(f"location {location} already taken.")
            for i, state2 in enumerate(board.states):
                if BoardState.compare(state, state2):
                    board.states.pop(i)
            board.states.append(BoardState(space_id, location, texture, tag))
            self.__state: BoardState = board.states[-1]
        def fetch(self) -> BoardState:
            return self.__state
        @staticmethod
        def checkLocationValidity(board: "Board", location: LocationArray) -> bool | str:
            return not any([
                int(location.x) > int(board.size.x),
                int(location.y) > int(board.size.y),
                int(location.x) <= 0,
                int(location.y) <= 0
            ])
        @staticmethod
        def ensureLocationValidity(board: "Board", location: LocationArray) -> None:
            if not Board.StatePlacementOperator.checkLocationValidity(board, location):
                raise ValueError(f"location {location} is out of bounds.")
    def lookFor(self, formula: str, **vars: Any) -> Any:
        for i, state in enumerate(self.states):
            if eval(formula, globals() | locals() | vars, {}):
                return i, state
        return 0, None

    def rmByState(self, state: BoardState) -> None:
        i, state = self.lookFor(f"state.space_id == space_id", space_id=state.space_id)
        if state is None:
            raise ValueError(f"tile with id '{space_id}' not found.")
        self.states.pop(i)

    # by id:

    def stateById(self, space_id: int) -> BoardState:
        return self.lookFor(f"state.space_id == space_id", space_id=space_id)[1]
    def locationById(self, space_id: int) -> LocationArray:
        return self.stateById(space_id).location
    def tagById(self, space_id: int) -> LocationArray:
        return self.stateById(space_id).tag
    def textureById(self, space_id: int) -> LocationArray:
        return self.stateById(space_id).texture
    def rmById(self, space_id: int) -> None:
        self.rmByState(self.stateById(space_id))

    # by tag:

    def stateByTag(self, tag: str) -> BoardState:
        return self.lookFor(f"state.tag == tag", tag=tag)[1]
    def locationByTag(self, tag: str) -> LocationArray:
        return self.stateByTag(tag).location
    def idByTag(self, tag: str) -> int:
        return self.stateByTag(tag).space_id
    def textureByTag(self, tag: str) -> LocationArray:
        return self.stateByTag(tag).texture
    def rmByTag(self, tag: str) -> None:
        self.rmByState(self.stateByTag(tag))

    # by location:

    def stateByLocation(self, location: LocationArray) -> BoardState:
        return self.lookFor(f"LocationArray.compare(state.location, location)", location=location)[1]
    def idByLocation(self, location: LocationArray) -> int:
        return self.stateByLocation(location).space_id
    def tagByLocation(self, location: LocationArray) -> str:
        return self.stateByLocation(location).tag
    def textureByLocation(self, location: LocationArray) -> LocationArray:
        return self.stateByLocation(location).texture
    def rmByLocation(self, location: LocationArray) -> None:
        self.rmByState(self.stateByLocation(location))

    def create(self, tag: str, name: str | None = None) -> object:
        """name parameter defaults to tag"""
        name = tag if name is None else name
        space_cls: type = ScriptHandler.createClass(name)
        space_id: int = self.generateRandomSpaceId()
        class new_space_cls(space_cls.__class__):
            id = space_id
            def rm(self1):
                self.rmById(space_id)
        space: new_space_cls = new_space_cls() # NOQA
        location_x: int = space.X
        location_y: int = space.Y
        texture: str = space.T
        color: str | ColorObject = space.C
        self.StatePlacementOperator(
            self,
            space_id,
            LocationArray(
                LocationX(location_x),
                LocationY(location_y)
            ),
            tag,
            TileTexture(
                texture,
                ColorObject(color) if isinstance(color, str) else color
            )
        )
        return space
    def render(self) -> str:
        lns = []
        for y in range(-int(self.size.y), 0):
            ln = [
                str(self.bg)
                if not self.stateTaken(LocationArray.from_int(x, -y))
                else str(self.stateTaken(LocationArray.from_int(x, -y)).texture)
                for x in range(1, int(self.size.x) + 1)
            ]
            lns.append("".join(ln))
        return "\n".join(lns)


class _Location:
    SIGN = ""
    def __init__(self, value: int) -> None:
        self.value: int = value
    def __int__(self) -> int:
        return self.value
    def __str__(self) -> str:
        return f"{self.SIGN}{self.value}"
class LocationX(_Location):
    SIGN = "x"
class LocationY(_Location):
    SIGN = "y"
class LocationArray:
    def __init__(self, x: LocationX, y: LocationY) -> None:
        self.x: LocationX = x
        self.y: LocationY = y
    def __iter__(self) -> iter:
        return iter([self.x, self.y])
    def __str__(self) -> str:
        return f"{self.x}{self.y}"
    @staticmethod
    def compare(location1: "LocationArray", location2: "LocationArray") -> bool:
        return str(location1) == str(location2)
    @classmethod
    def from_int(cls, x: int, y: int) -> "LocationArray":
        return cls(LocationX(x), LocationY(y))
class LocationOffset:
    def __init__(self, x_offset: LocationX, y_offset: LocationY) -> None:
        self.x: LocationX = x_offset
        self.y: LocationY = y_offset
    @classmethod
    def generate(cls, key: str) -> "LocationOffset":
        try:
            return cls(LocationX({
                "q": -1, "w": +0, "e": +1,
                "a": -1, "s": +0, "d": +1,
                "z": -1, "x": +0, "c": +1
            }[key]), LocationY({
                "q": +1, "w": +1, "e": +1,
                "a": +0, "s": -1, "d": +0,
                "z": -1, "x": -1, "c": -1
            }[key]))
        except KeyError:
            raise ValueError(f"invalid direction: '{key}'")
    def apply(self, location: LocationArray) -> LocationArray:
        return LocationArray(
            LocationX(
                int(location.x) + int(self.x)
            ),
            LocationY(
                int(location.y) + int(self.y)
            )
        )
    def __iter__(self) -> iter:
        return iter([int(self.x), int(self.y)])

class DirectionObject:
    def __init__(self, key: str) -> None:
        key = key.lower().replace("_", "-")
        try:
            self.__key: str = {
                "up-left": "q",
                "up": "w",
                "up-right": "e",
                "left": "a",
                "right": "d",
                "down-left": "z",
                "down": "x",
                "down-right": "c"
            }[key]
        except KeyError:
            raise ValueError(f"invalid direction: '{key}'")
        self.__name: str = key
        self.__arrow: str = {
            "q": "⬉",
            "w": "⬆",
            "e": "⬈",
            "a": "⬅",
            "s": "⬇",
            "d": "⮕",
            "z": "⬋",
            "x": "⬇",
            "c": "⬊"
        }[self.__key]
        self.__x_offset, self.__y_offset = LocationOffset.generate(self.__key)
    def to_offset(self) -> tuple[int]:
        return self.__x_offset, self.__y_offset
    def to_arrow(self) -> str:
        return self.__arrow
    def to_key(self) -> str:
        return self.__key
    def __str__(self) -> str:
        return self.__name

class _DirectionClass:
    def __getattr__(self, key: str) -> DirectionObject:
        return DirectionObject(key)

Direction: _DirectionClass = _DirectionClass()
ALL_POSSIBLE_DIRECTIONS: list[DirectionObject] = [
    Direction.up_left,
    Direction.up,
    Direction.up_right,
    Direction.left,
    Direction.right,
    Direction.down_left,
    Direction.down,
    Direction.down_right
]

class Route:
    def __init__(self, from_: tuple[int, int], to: tuple[int, int] | str) -> None:
        self.from_: list[int] = list(from_)
        self.to: list[int] = list(to)
    def calculateOptimalDirection(self, use_qezc: bool = True, return_as_str: bool = False) -> Direction:
        x_offset: int = 0
        y_offset: int = 0
        if self.from_[0] < self.to[0]:
            x_offset += 1
        if self.from_[0] > self.to[0]:
            x_offset -= 1
        if self.from_[1] < self.to[1]:
            y_offset += 1
        if self.from_[1] > self.to[1]:
            y_offset -= 1

        for direction in ALL_POSSIBLE_DIRECTIONS:
            if LocationArray.compare(LocationArray(x_offset, y_offset), LocationArray(*direction.to_offset())):
                return direction
    def updateLocation(self, x: int, y: int) -> None:
        self.from_[0] = x
        self.from_[1] = y
    def updateTarget(self, x: int, y: int) -> None:
        self.to[0] = x
        self.to[1] = y
    def __str__(self) -> str:
        return f"x{self.from_[0]}y{self.from_[1]} -> x{self.to[0]}y{self.to[1]}"


import time

from typing import Callable, Any
from colorama import Style, Fore

from .essentials import ScriptHandler

# economy:

class Currency:
    def __init__(self, name: str, sign: str, color: "ColorObject") -> None:
        self.name: str = name
        self.sign: str = sign
        self.color: "ColorObject" = color
    @classmethod
    def setup(cls, name: str) -> "Currency":
        sh: ScriptHandler = ScriptHandler()
        currency: object = sh.createClass(name)
        return cls(name, currency.S, currency.C)
    def apply(self, number: float, *, name: bool = False, sign: bool = True, color: bool = True) -> str:
        return (self.color.apply if color else str)(f"{self.sign if sign else ''}{number}{' '+self.name if name else ''}")

class TaxSystem:
    def __init__(self, name: str, description: str, amount: str, color: "ColorObject", tp: int = 0, hidden: bool = False, every_x_ticks: int = 1, times_effective: int = -1):
        self.name: str = name
        self.description: str = description
        self.amount: str = f"round({amount})"
        self.color: "ColorObject" = color
        self.tp: int = tp
        self.hidden: bool = hidden
        self.every_x_ticks: int = every_x_ticks
        self.times_effective: int = times_effective
        self.times_unpaid: int = 0
    @classmethod
    def setup(cls, name: str) -> "TaxSystem":
        sh: ScriptHandler = ScriptHandler()
        tax_system: object = sh.createClass(name)
        return cls(name, tax_system.D, tax_system.A, tax_system.C, tax_system.H, tax_system.X, tax_system.F)
    def whenUnpaidDno(self) -> None: # dno = do not overwrite (in class children)
        self.times_effective += 1
        self.unpaid()
    def unpaid(self) -> None:
        print(f"{Fore.RED}insufficient funds!{Style.RESET_ALL}")
        exit(1)

class OneTimePayment(TaxSystem):
    def __init__(self, listing: "ShopItemListing") -> None:
        super().__init__(f"purchased {listing.item.name}", listing.item.description, listing.price.amount, listing.color, times_effective=1)

class Wallet:
    def __init__(self, currency: Currency, amount: float, taxes: list[TaxSystem] | None = None, tp: int = 0) -> None:
        self.currency: Currency = currency
        self.amount: float = amount
        self.taxes: list[list[TaxSystem, int, int]] = [] # [tax, tick, times_applied]
        self.tp: int = tp
        if taxes:
            for tax in taxes:
                self.addTax(tax)
    @classmethod
    def setup(cls, name: str) -> "Wallet":
        sh: ScriptHandler = ScriptHandler()
        wallet: object = sh.createClass(name)
        return cls(Currency.setup(wallet.V), wallet.A, [TaxSystem.setup(i) for i in wallet.T])
    def deposit(self, amount: float) -> None:
        self.amount += amount
    def withdraw(self, amount: float) -> bool:
        if amount <= self.amount:
            self.amount -= amount
            return True
        else:
            return False
    def round(self) -> None:
        self.amount = round(self.amount)
    def tick(self, *, delay: float = 1.0) -> None:
        remove: list[int] = []
        for i, (tax, tick, times_applied) in enumerate(self.taxes):
            if tax.tp < self.tp or (tax.tp != -1 and self.tp == -1):
                continue
            if tax.times_effective == times_applied:
                remove.append(i)
                continue
            if tick == tax.every_x_ticks:
                self.tax(tax, delay=delay)
                self.taxes[i][1] = 1
                self.taxes[i][2] += 1
            else:
                self.taxes[i][1] += 1
        for i in remove:
            self.taxes.pop(i)
    def tax(self, tax: TaxSystem, *, delay: float = 1.0) -> None:
        x = self.amount
        amount = eval(tax.amount)
        if self.amount < amount:
            raise ValueError("insufficient funds")
        self.withdraw(amount)
        tick = 1
        if not tax.hidden:
            print(tax.color.apply(f"-{self.currency.apply(amount, color=False)}, ({tax.name}: {tax.description})"))
            time.sleep(delay)
    def addTax(self, tax: TaxSystem) -> None:
        self.taxes.append([tax, 1, 0])
    def __str__(self, **kwargs: Any) -> str:
        return self.currency.apply(self.amount, **kwargs)

# item:

class Item:
    def __init__(self, name: str, description: str, effect: Callable | str) -> None:
        self.name: str = name
        self.description: str = description
        self.effect: Callable = lambda _: print(effect) if isinstance(effect, str) else effect
    @classmethod
    def setup(cls, name: str) -> "Item":
        sh: ScriptHandler = ScriptHandler()
        item: object = sh.createClass(name)
        return cls(listing.N, listing.D, listing.E)
    def use(self, game: "Game", *, delay: float = 2.0) -> None:
        self.effect(game)
        time.sleep(delay)
class ItemInventory:
    def __init__(self, wallet: Wallet, items: list[Item] | None = None) -> None:
        self.wallet: Wallet = wallet
        self.items: list[Item] = items or []
    @classmethod
    def setup(cls) -> "ItemInventory":
        sh: ScriptHandler = ScriptHandler()
        player: object = sh.createClass("Player")
        return cls(Wallet.setup(player.W), [Item.setup(i) for i in player.I])
    def get(self, name: str, if_not_found: Callable | None = None) -> Item:
        for item in self.items:
            if item.name.lower().replace("_", " ") == name.lower().replace("_", " "):
                return item
        if if_not_found:
            return if_not_found
        raise ValueError(f"no item named '{name}' found in the inventory.")
    def add(self, item: Item, *, delay: float = 1.0) -> None:
        print(f"{Fore.MAGENTA}new item received: {item.name}{Style.RESET_ALL}")
        time.sleep(delay)
        self.items.append(item)
    def remove(self, item: Item) -> None:
        for i, l_item in enumerate(self.items):
            if l_item.name == item.name:
                self.items.pop(i)
                return
class ShopItemListing:
    def __init__(self, item: Item, price: Wallet, color: "ColorObject") -> None:
        self.item: Item = item
        self.price: Wallet = price
        self.color: "ColorObject" = color
    @classmethod
    def createItem(cls, name: str, description: str, effect: Callable, price: float, currency: Currency, color: "ColorObject") -> "ShopItemListing":
        return cls(Item(name, description, effect), Wallet(currency, price), color)
    @classmethod
    def fromFunction(cls, price: float, currency: Currency, color: "ColorObject") -> Callable:
        def wrapper1(function: Callable) -> "ShopItemListing":
            return cls.createItem(function.__name__.replace("_", " "), function.__doc__, function, price, currency, color)
        return wrapper1
    @classmethod
    def setup(cls, name: str) -> "ShopItemListing":
        sh: ScriptHandler = ScriptHandler()
        listing: object = sh.createClass(name)
        return cls.createItem(listing.N if hasattr(listing, "N") else name, listing.D, listing.E, listing.P, Currency.setup(listing.V), listing.C)

    def __str__(self) -> str:
        return (f"{self.color.apply(self.item.name)}\n"
                f"- {'\n- '.join(self.item.description.split('\n'))}\n"
                f"- costs: {self.price}")
    def buy(self, inventory: ItemInventory, amount: int = 1, *, delay: float = 1.0) -> None:
        otp: OneTimePayment = OneTimePayment(self)
        try:
            inventory.wallet.tax(otp, delay=delay)
        except ValueError:
            otp.when_unpaid()
            time.sleep(delay)
            return
        inventory.add(self.item, delay=delay)

class Shop:
    def __init__(self, name: str, color: "ColorObject", items: list[ShopItemListing]) -> None:
        self.name: str = name
        self.color: "ColorObject" = color
        self.items: list[ShopItemListing] = items
    @classmethod
    def setup(cls, name: str) -> "Shop":
        sh: ScriptHandler = ScriptHandler()
        shop: object = sh.createClass(name)
        return cls(name, shop.C, [ShopItemListing.setup(i) for i in shop.I])
    def display(self, *, delay: float = 1.0) -> ShopItemListing:
        print(self.color.apply(f"{self.name}:"))
        for i, item in enumerate(self.items, start=1):
            print(f"{Fore.GREEN}{i}. {Style.RESET_ALL}{item}")
        while True:
            try:
                choice: str | int = input("choose an item (1-" + str(len(self.items)) + ") or 'c' to cancel: ").lower()
                if choice == "c":
                    print(f"{Fore.LIGHTBLACK_EX}operation cancelled{Style.RESET_ALL}")
                    time.sleep(delay)
                    return
                choice = int(choice)
            except ValueError:
                print(f"{Fore.RED}invalid choice{Style.RESET_ALL}")
                continue
            if 1 <= choice <= len(self.items):
                return self.items[choice-1]
            else:
                print(f"{Fore.RED}invalid choice{Style.RESET_ALL}")
    def send(self, game: "Game", command: str = "m", enabled: bool = True) -> "Shop":
        game.newShop(self, command)
        game.toggleShop(self, enabled)
        return self

def newItem(name: str, shop_listing: bool = True) -> ShopItemListing:
    item: ShopItemListing = ShopItemListing.setup(name)
    return item if shop_listing else item.item


from difflib import get_close_matches

from .location import *
from .essentials import *

class ColorObject:
    def __init__(self, color: str) -> None:
        color = str(color).lower()
        if color in ["grey", "gray"]:
            color = "LIGHTBLACK_EX"
        if color == "brown":
            color = "ansi38m2m139m69m19m"
        if color == "reset":
            color = str(Style.RESET_ALL)
        elif color.startswith("ansi") and color.endswith("m"):
            ansi_codes: list[str] = color.removeprefix("ansi").removesuffix("m").split("m")
            self.color = eval(rf"'\033[{';'.join(ansi_codes)}m'")
        else:
            color = color.upper()
            if not hasattr(Fore, color):
                closest: list[str] = get_close_matches(color, Fore.__dict__)
                raise ValueError(f"no color named {color}.{' did you mean: \''+closest[0]+'\'' if closest else ''}")
            self.color: str = str(getattr(Fore, color))
    def apply(self, string: str) -> str:
        return f"{self.color}{string}{Style.RESET_ALL}"
class _ColorClass:
    def __getattr__(self, item: str) -> ColorObject:
        return ColorObject(item)
Color: _ColorClass = _ColorClass()
class TileTexture:
    def __init__(self, texture: str, color: ColorObject) -> None:
        if len(texture) > 1:
            raise ValueError("texture must be only 1 character.")
        self.texture: str = str(texture)
        self.color: ColorObject = color
    def __str__(self, color: bool = True) -> str:
        return self.color.apply(self.texture) if color else self.texture
class BoardState:
    def __init__(self, space_id: int, location: LocationArray, texture: TileTexture, tag: str) -> None:
        self.space_id: int = space_id
        self.location: LocationArray = location
        self.texture: TileTexture = texture
        self.tag: str = tag
    def __str__(self) -> str:
        return f"{self.space_id} ({self.texture}) at {self.location} with tag '{self.tag}'"
    @staticmethod
    def compare(state1: "BoardState", state2: "BoardState") -> bool:
        return str(state1) == str(state2)
class Piston:
    def __init__(self, board: "Board") -> None:
        self.board: "Board" = board
        self.pushes: int = 0
    def push(self,
        space_id: int,
        key_or_offset: str | LocationOffset,
        distance: int = 1,
        destructive: bool | None = False,
        cannot_go_there_warning: bool = False
    ) -> LocationArray:
        pushes_before: int = self.pushes
        for _ in range(distance):
            offset: LocationOffset = key_or_offset if isinstance(key_or_offset, LocationOffset) else LocationOffset.generate(key_or_offset)
            old_location: LocationArray = self.board.locationById(space_id)
            new_location: LocationArray = offset.apply(old_location)
            if cannot_go_there_warning and not self.board.StatePlacementOperator.checkLocationValidity(self.board, new_location):
                break
            self.board.StatePlacementOperator.ensureLocationValidity(self.board, new_location)
            if self.board.stateTaken(new_location):
                if destructive is True:
                    ...
                elif destructive is False:
                    break
                elif destructive is None:
                    raise ValueError("non-destructive push attempt on occupied location.")
            self.board.StatePlacementOperator(
                self.board,
                space_id,
                new_location,
                self.board.tagById(space_id),
                self.board.textureById(space_id)
            )
            self.board.rmById(space_id)
            self.pushes += 1
        pushes_after: int = self.pushes
        elapsed_pushes: int = pushes_after - pushes_before
        if cannot_go_there_warning and elapsed_pushes == 0:
            print(f"{Fore.RED}cannot go there!{Style.RESET_ALL}")
            time.sleep(0.4)
        return new_location # NOQA
class _SmClass:
    def __matmul__(self, other: int | float) -> tuple[int, int]:
        return int(other), int(other / 2)

Sm = _SmClass()


from typing import Callable, Any

from .essentials import ScriptHandler
from .location import Route
from .space import Piston


class Bot:
    SERVICES: list[Any] = []
    def __init__(self, game: "Game", tag: str, piston: "Piston", route: Route, speed: int = 1) -> None:
        self.game: "Game" = game
        self.tag: str = tag
        self.piston: "Piston" = piston
        self.route: Route = route
        self.speed: int = speed
        self.completed: bool = False
        for service in self.SERVICES:
            service(self)
    @classmethod
    def setup(cls, game: "Game", tag: str) -> "SimpleBot":
        sh: ScriptHandler = ScriptHandler()
        bot: object = sh.createClass(tag)
        return cls(game, tag, Piston(game.board), Route((bot.X, bot.Y), bot.P), bot.S)
    @classmethod
    def withService(cls, obj: Any) -> None:
        class _Bot(cls):
            SERVICES = cls.SERVICES + [obj]
        return _Bot
    def tick(self, destructive_push: bool | None = False, complete_on_failed_calculation: bool = True, complete_on_failed_push: bool = False) -> None:
        if self.completed:
            return
        direction = self.route.calculateOptimalDirection()
        if direction is None:
            self.completed = complete_on_failed_calculation
            return
        location: "LocationArray | None" = self.piston.push(self.game.board.idByTag(self.tag), direction.to_key(), self.speed, destructive_push)
        if location is None:
            self.completed = complete_on_failed_push
            return
        self.route.updateLocation(int(location.x), int(location.y))
def simpleTracker(tag: str) -> Callable:
    def wrapper(bot: Bot) -> None:
        old_tick_method: Callable = bot.tick
        update_location: Callable = lambda: bot.route.updateTarget(*map(int, bot.game.board.locationByTag(tag)))
        def new_tick_method(destructive_push: bool | None = False, complete_on_failed_push: bool = False) -> None:
            update_location()
            old_tick_method(destructive_push, False, complete_on_failed_push)
        update_location()
        bot.tick = new_tick_method
    return wrapper


from setuptools import setup, find_packages

def read_long_description():
    with open('README.md', encoding='utf-8') as f:
        return f.read()

setup(
    name='brary',
    version='1',
    author='fossil',
    author_email='fossil.org1@gmail.com',
    description='byte-sized script-based 2d art and game creation framework.',
    long_description=read_long_description(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.12',
    install_requires=["colorama"],
)